#include "../include/parse.h"
#include "../include/constr.h"
#include "../include/eye_log.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <netdb.h> 
#include <unistd.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <sys/socket.h>

#include <sys/ipc.h>
#include <sys/shm.h>

typedef unsigned char       uchar;
typedef unsigned char       byte;
typedef unsigned short int  word;
typedef unsigned long int   dword;

#define ICMP_ECHO 8
#define ICMP_ECHOREPLY 0
#define TIMEOUT_CMD "timeout"
#define COUNTER_CMD "startcounter"
#define SENDTIME_CMD "sendtime"

typedef struct _ip
{
    byte ip_verlen;
    byte ip_tos;
    word ip_len;
    word ip_id;
    word ip_fragoff;
    byte ip_ttl;
    byte ip_proto;
    word ip_chksum;
    struct in_addr ip_src;
    struct in_addr ip_dst;
} ip;

typedef struct _icmp
{
    byte icmp_type;
    byte icmp_code;
    word icmp_chksum;
    word icmp_id;
    word icmp_seq;
} icmp;

void sendping(int s, struct sockaddr_in *sa_local, int count);
word GetSum(word *buf, int len);
void reset(Plugin * pl, char *mem, int i);
void set(Plugin * pl, char *mem, int i, int status, int cell)
;

static int TimeOut=3e4;			// in msek
static int StartCounter=10;		
static int SendTime=1e3;		// in msek

enum { PE_SECTIONABSENT=1, PE_MEMORY, PE_OPENSOCKET, PE_NAMERESOLVE};

int main(int argv, char **argc)
{
    struct hostent *he;
    int len, count, i, j, k;
    byte data[256];
    icmp *icmph;
    ip *iph;
    struct timeval tv, last;
    struct timeval *tvold;
    CELL_M *status;

    Plugin *pl;
    PutSection *ps;
    Host *h;
    Command *c;

    int s;
    struct sockaddr_in *sa_local;
    struct sockaddr_in sa_host;
    int sa_len;

    char * memaddr;

    if(argv < 2) 
    {
	printf("missing parametr\n");
	return 1;
    }
    if(OpenCfgFile(argc[1], 0)!=OCF_OK)
    {
	printf("cannot open cfg file\n");
	return 1;
    }

    {
	int memhnd;
        key_t key;  
	long len;
        key=ftok((argv>2) ? argc[2] : "/var/tmp", 'e');
        if(key==-1) { perror("Ftok"); exit(1); }
	memhnd=shmget(key, sizeof(long), 0660);
	if(memhnd<0) { perror("shmget1"); exit(1); }
        memaddr=shmat(memhnd, NULL, 0);
	if(memaddr<0) { perror("shmat1"); exit(1); }
        len=*((long *)memaddr);
	if((shmdt(memaddr))<0) { perror("shmdt"); exit(1); }
        memhnd=shmget(key, len, 0660);
	if(memhnd<0) { perror("shmget2"); exit(1); }
        memaddr=shmat(memhnd,NULL,0);
	if(memaddr<0) { perror("shmat2"); exit(1); }
    }

    pl=GetFirstPlugin();
    status=GetStatusByName(pl->NameOfCfg, memaddr);
    status->status=0;
    count=0;
    for(ps=pl->ps_next; ps; ps=ps->ps_next)
	for(h=ps->h_next; h; h=h->h_next) count++;
    if(count==0) 
    {
	status->status=1;
	status->cell=PE_SECTIONABSENT;
	exit(2); 
    }
    

    sa_local=(struct sockaddr_in *)calloc(count, sizeof(struct sockaddr_in));
    if(!sa_local)
    {
	printf("error allocation memory\n");
	status->status=1;
	status->cell=PE_MEMORY;
	return 1;
    }

    for(c=pl->p_cmd; c; c=c->c_next)
    {
	if(!strcasecmp(c->c_cmd, TIMEOUT_CMD)) 
	{	
	    TimeOut=atoi(c->c_value);
	}else if(!strcasecmp(c->c_cmd, COUNTER_CMD)) 
	{	
	    StartCounter=atoi(c->c_value);
	}else if(!strcasecmp(c->c_cmd, SENDTIME_CMD)) 
	{	
	    SendTime=atoi(c->c_value);
	}
    }
    
    s=socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
    if(s<0)
    {
	printf("error open socket\n");
	status->status=1;
	status->cell=PE_OPENSOCKET;
	return 1;
    }
    i=0;
    for(ps=pl->ps_next; ps; ps=ps->ps_next) for(h=ps->h_next; h; h=h->h_next)
    {
	memset((byte *)&sa_local[i].sin_addr, 0, 4);
	
	he=gethostbyname(h->_h_addr);
	if(!he)
	{
	    sa_local[i].sin_family=!AF_INET;
	    set(pl, memaddr, i, PE_NAMERESOLVE, PE_NAMERESOLVE);
	}else
	{
	    sa_local[i].sin_family=AF_INET;
	    memcpy( (byte *)&sa_local[i].sin_addr, (byte *)he->h_addr, 4);
	    set(pl, memaddr, i, 0, StartCounter);
	}
	i++;
    }

    sa_len=sizeof(struct sockaddr_in);
    sa_host.sin_family=AF_INET;
    
    while(brk)
    {
        int n;
        struct timeval timeout, now, intvl;
        fd_set rfds;
	
    	sendping(s, sa_local, count);
    	gettimeofday(&last, NULL);
	
	while(1)
	{
	    FD_ZERO(&rfds);
	    FD_SET(s, &rfds);
	    
	    intvl.tv_sec=0;
	    intvl.tv_usec=SendTime*1e3;
	    while (timeout.tv_usec >= 1e6) 
	    {
		timeout.tv_usec -= 1e6;
		timeout.tv_sec++;
	    }

	    gettimeofday(&now, NULL);
	    timeout.tv_sec = last.tv_sec + intvl.tv_sec - now.tv_sec;
	    timeout.tv_usec = last.tv_usec + intvl.tv_usec - now.tv_usec;
	    while(timeout.tv_usec < 0) 
	    {
		timeout.tv_usec += 1e6;
		timeout.tv_sec--;
	    }
	    while (timeout.tv_usec >= 1e6) 
	    {
		timeout.tv_usec -= 1e6;
		timeout.tv_sec++;
	    }

	    if (timeout.tv_sec < 0) 
	    {
		timeout.tv_sec = timeout.tv_usec = 0;
		break;
	    }

	    n = select(s+1, &rfds, NULL, NULL, &timeout);
	    if(n >= 1) 
	    do {
		if((len=recvfrom(s, data, sizeof(data), 0, 0, 0/*&sa_host, &sa_len*/)) == -1) continue;

    		gettimeofday(&tv, 0);
		iph = (ip *)data;
		if(len < 4*(iph->ip_verlen&0x0f) + sizeof(icmp)) continue;
    		icmph=(icmp *)(data + 4*(iph->ip_verlen&0x0f));
            	
		if(icmph->icmp_type != ICMP_ECHOREPLY) continue;
		if(icmph->icmp_id != getpid()) continue;

		k=*((int *)(((byte *)icmph+sizeof(icmp))))
;
    		tvold=(struct timeval *)((byte *)icmph+sizeof(icmp)+sizeof(int));

		j=(tv.tv_sec-tvold->tv_sec)*1e6 + (tv.tv_usec-tvold->tv_usec);
		if(TimeOut!=0 && TimeOut*1e3<j) continue;
	
#ifdef DEBUG
		printf("PING: --- received from %i.%i.%i.%i\n", (sa_local[k].sin_addr.s_addr)&0xff,
                                            (sa_local[k].sin_addr.s_addr>>8)&0xff, 
					    (sa_local[k].sin_addr.s_addr>>16)&0xff, 
					    (sa_local[k].sin_addr.s_addr>>24)&0xff);
#endif
		reset(pl, memaddr, k);
	    } while(0);

    	}
 	sleep(SendTime);
   }

    free(sa_local);
    return 0;
}

void sendping(int s, struct sockaddr_in *sa_local, int count)
{
    icmp *icmph;
    int len, sent_len;
    int sa_len, i;
    int *num;
    byte data[128];
    struct timeval *tv;
    
    icmph=(icmp *)data;
    len = sizeof(icmp) + sizeof(int) + sizeof(struct timeval);
    num=(int *)((byte *)icmph+sizeof(icmp));
    tv=(struct timeval *)((byte *)icmph+sizeof(icmp)+sizeof(int));

    for(i=0; i<count; i++)
    {
	if(sa_local[i].sin_family != AF_INET) continue;
	memset(data, 0, 128);
	icmph->icmp_type = ICMP_ECHO;
	icmph->icmp_code = 0;
	icmph->icmp_id = getpid();
	icmph->icmp_seq = 0;
	icmph->icmp_chksum = 0;
	sa_len = sizeof(struct sockaddr_in);
	*num=i;
    	gettimeofday(tv, 0);

    	icmph->icmp_chksum = GetSum( (word *)data, len);
    	sent_len = sendto(s, data, len, 0, &sa_local[i], sa_len);

    	if(sent_len == -1) continue;
    	if(sent_len != len) continue;
#ifdef DEBUG
	printf("PING: --- sended to %x\n", htonl(sa_local[i].sin_addr.s_addr));
#endif
    }
}

word GetSum(word *buf, int len)
{
    dword sum=0;
    
    while(len > 1)
    {
	sum += *buf++;
	len -= 2;
    }
    
    if(len == 1)
	sum += *((byte *)buf);
	
    sum = (sum>>16) + (sum&0xffff);
    return (word)~sum;
}

void reset(Plugin * pl, char *mem, int i)
{
    PutSection *ps ;
    Host *h;
    CELL_M *ce;
    short *sh,*shp;
    int j=0;
    
    sh=GetCfgByName(pl->NameOfCfg, mem);
    ps=pl->ps_next;
    while(ps && j<=i)
    {
	shp=GetPutByName(ps->ps_name, sh);
	if(!shp) return;
	for(h=ps->h_next; h; j++, h=h->h_next)
	{
	    ce=GetCELL_M(shp);
	    if(i==j) 
	    {
		ce->cell=StartCounter;
		if(ce->status!=0)writedblog(pl->NameOfCfg,h->h_key,0);
		ce->status=0;
		j++;
		break;
	    }
	}
	ps=ps->ps_next;
    } 
}

void set(Plugin * pl, char *mem, int i, int status, int cell)
{
    PutSection *ps ;
    Host *h;
    CELL_M *ce;
    short *sh,*shp;
    int j=0;
    
    sh=GetCfgByName(pl->NameOfCfg, mem);
    ps=pl->ps_next;
    while(ps && j<=i)
    {
	shp=GetPutByName(ps->ps_name, sh);
	if(!shp) return;
	for(h=ps->h_next; h; j++, h=h->h_next)
	{
	    ce=GetCELL_M(shp);
	    if(i==j) 
	    {
		if(ce->status!=status)writedblog(pl->NameOfCfg,h->h_key,status);
		ce->status=status;
		ce->cell=cell;
		j++;
		break;
	    }
	}
	ps=ps->ps_next;
    } 
}
